import os
import tempfile

import tensorflow as tf
from moviepy import VideoFileClip
from moviepy.video.compositing import CompositeVideoClip
from moviepy.video.fx import AccelDecel, TimeSymmetrize
from moviepy.video.VideoClip import DataVideoClip

from dreamify.utils.common import deprocess_image


class ImageToVideoConverter:
    def __init__(self, dimensions, max_frames_to_sample):
        self.dimensions = dimensions
        self.current_chunk: list = []
        self.max_frames_to_sample: int = max_frames_to_sample
        self.curr_frame_idx: int = 0

        self.FPS: int = 30
        self.MAX_FRAMES_IN_MEM: int = 50

        self.chunk_files: list = []
        self.temp_folder = tempfile.mkdtemp()
        print(f"Temporary folder created at {self.temp_folder}")

    def add_to_frames(self, frame):
        frame = tf.image.resize(frame, self.dimensions)
        frame = deprocess_image(frame)
        self.current_chunk.append(frame)
        self.curr_frame_idx += 1

        if len(self.current_chunk) >= self.MAX_FRAMES_IN_MEM:
            self.flush_chunk()

    def continue_framing(self):
        return self.curr_frame_idx < self.max_frames_to_sample - 1

    def flush_chunk(self):
        if not self.current_chunk:
            return

        chunk_frames = []
        for i in range(len(self.current_chunk) - 1):
            chunk_frames.append(self.current_chunk[i])
            interpolated = self.interpolate_frames(
                tf.cast(self.current_chunk[i], tf.float32),
                tf.cast(self.current_chunk[i + 1], tf.float32),
                tf.constant(30),
            )
            chunk_frames.extend(interpolated)

        chunk_frames.append(self.current_chunk[-1])

        clip = DataVideoClip(self.current_chunk, lambda x: x, fps=self.FPS)
        chunk_path = os.path.join(
            self.temp_folder, f"chunk_{len(self.chunk_files)}.mp4"
        )
        clip.write_videofile(chunk_path, logger=None)
        clip.close()

        self.chunk_files.append(chunk_path)
        self.current_chunk = []

    def to_video(
        self,
        output_path="dream.mp4",
        duration=3,
        mirror_video=False,
    ):
        self.flush_chunk()

        clips = [VideoFileClip(chunk) for chunk in self.chunk_files]

        final_clip = CompositeVideoClip.concatenate_videoclips(clips)

        final_clip = AccelDecel(new_duration=duration).apply(final_clip)

        if mirror_video:
            final_clip = TimeSymmetrize().apply(final_clip)

        final_clip.write_videofile(output_path, logger=None)
        final_clip.close()

    def to_gif(
        self,
        output_path="dream.gif",
        duration=3,
        mirror_video=False,
    ):
        self.flush_chunk()

        clips = [VideoFileClip(chunk) for chunk in self.chunk_files]

        final_clip = CompositeVideoClip.concatenate_videoclips(clips)

        final_clip = AccelDecel(new_duration=duration).apply(final_clip)

        if mirror_video:
            final_clip = TimeSymmetrize().apply(final_clip)

        final_clip.write_gif(output_path, fps=30, logger=None)
        final_clip.close()

    @tf.function
    def interpolate_frames(self, frame1, frame2, num_frames):
        alphas = tf.linspace(0.0, 1.0, num_frames + 2)[1:-1]
        interpolated_frames = (1 - alphas[:, None, None, None]) * frame1 + alphas[
            :, None, None, None
        ] * frame2
        return tf.cast(interpolated_frames, tf.uint8)



# import IPython.display as display
from pathlib import Path

import tensorflow as tf

from dreamify.lib import Config, FeatureExtractor, TiledGradients, validate_dream_params
from dreamify.utils import (
    deprocess_image,
    get_image,
    preprocess_image,
    save_output,
    show,
)


@validate_dream_params
def deepdream(
    image_path,
    output_path="dream.png",
    iterations=100,
    learning_rate=0.01,
    octaves=range(-2, 3),
    octave_scale=1.3,
    save_video=False,
    save_gif=False,
    duration=3,
    mirror_video=False,
):
    output_path = Path(output_path)

    # base_model = tf.keras.applications.InceptionV3(
    #     include_top=False, weights="imagenet"
    # )

    # names = ["mixed3", "mixed5"]
    # layers = [base_model.get_layer(name).output for name in names]

    # ft_ext = tf.keras.Model(inputs=base_model.input, outputs=layers)
    # get_tiled_gradients = TiledGradients(ft_ext)

    # deepdream_model = tf.keras.Model(inputs=base_model.input, outputs=layers)

    img = get_image(image_path)
    img = preprocess_image(img)

    original_shape = img.shape[1:-1]

    ft_ext = FeatureExtractor("inception_v3", dream_style="deep", layer_settings=None)
    get_tiled_gradients = TiledGradients(ft_ext.model)

    img = get_image(image_path)
    img = preprocess_image(img)

    original_shape = img.shape[1:-1]

    config = Config(
        feature_extractor=ft_ext,
        layer_settings=ft_ext.layer_settings,
        original_shape=original_shape,
        save_video=save_video,
        save_gif=save_gif,
        enable_framing=True,
        duration=duration,
        mirror_video=mirror_video,
        max_frames_to_sample=iterations * len(octaves),
    )

    for octave in octaves:
        new_size = tf.cast(tf.convert_to_tensor(original_shape), tf.float32) * (
            octave_scale**octave
        )
        new_size = tf.cast(new_size, tf.int32)
        img = tf.image.resize(img, tf.cast(new_size, tf.int32))

        for iteration in range(iterations):
            gradients = get_tiled_gradients(tf.squeeze(img), new_size)
            img = img + gradients * learning_rate
            img = tf.clip_by_value(img, -1, 1)

            if iteration % 10 == 0:
                # display.clear_output(wait=True)
                show(deprocess_image(img))
                print("Octave {}, Iteration {}".format(octave, iteration))

            if config.enable_framing and config.framer.continue_framing():
                config.framer.add_to_frames(img)

    img = deprocess_image(img)
    show(img)

    save_output(img, output_path, config)

    return img


def main(img_path, save_video=False, save_gif=False, duration=3, mirror_video=False):
    if img_path is None:
        img_path = (
            "https://storage.googleapis.com/download.tensorflow.org/"
            "example_images/YellowLabradorLooking_new.jpg"
        )

    deepdream(
        image_path=img_path,
        save_video=save_video,
        save_gif=save_gif,
        duration=duration,
        mirror_video=mirror_video,
    )


if __name__ == "__main__":
    main()



import warnings
from pathlib import Path

import tensorflow as tf

from dreamify.lib import Config, FeatureExtractor, validate_dream_params
from dreamify.utils import (
    deprocess_image,
    get_image,
    preprocess_image,
    save_output,
    show,
)
from dreamify.utils.dream_utils import gradient_ascent_loop

warnings.filterwarnings(
    "ignore", category=UserWarning, module="keras.src.models.functional"
)


@validate_dream_params
def dream(
    image_path,
    output_path="dream.png",
    model_name="inception_v3",
    learning_rate=20.0,
    iterations=30,
    octaves=3,
    octave_scale=1.4,
    max_loss=15.0,
    save_video=False,
    save_gif=False,
    duration=3,
    mirror_video=False,
):
    output_path = Path(output_path)

    ft_ext = FeatureExtractor(model_name, dream_style="shallow", layer_settings=None)

    original_img = get_image(image_path)
    original_img = preprocess_image(original_img)

    original_shape = original_img.shape[1:-1]

    config = Config(
        feature_extractor=ft_ext,
        layer_settings=ft_ext.layer_settings,
        original_shape=original_shape,
        save_video=save_video,
        save_gif=save_gif,
        enable_framing=save_video,
        duration=duration,
        mirror_video=mirror_video,
        max_frames_to_sample=iterations,
    )

    successive_shapes = [original_shape]
    for i in range(1, octaves):
        shape = tuple([int(dim / (octave_scale**i)) for dim in original_shape])
        successive_shapes.append(shape)
    successive_shapes = successive_shapes[::-1]

    shrunk_original_img = tf.image.resize(original_img, successive_shapes[0])

    img = tf.identity(original_img)
    for i, shape in enumerate(successive_shapes):
        print(
            f"\n\n{'_'*20} Processing octave {i + 1} with shape {successive_shapes[i]} {'_'*20}\n\n"
        )
        img = tf.image.resize(img, successive_shapes[i])
        img = gradient_ascent_loop(
            img,
            iterations=iterations,
            learning_rate=learning_rate,
            max_loss=max_loss,
            config=config,
        )
        upscaled_shrunk_original_img = tf.image.resize(
            shrunk_original_img, successive_shapes[i]
        )
        same_size_original = tf.image.resize(original_img, successive_shapes[i])
        lost_detail = same_size_original - upscaled_shrunk_original_img
        img += lost_detail
        shrunk_original_img = tf.image.resize(original_img, successive_shapes[i])

    img = deprocess_image(img)
    show(img)

    save_output(img, output_path, config)

    return img


def main(img_path, save_video=False, save_gif=False, duration=3, mirror_video=False):
    if img_path is None:
        img_path = (
            "https://storage.googleapis.com/download.tensorflow.org/"
            "example_images/YellowLabradorLooking_new.jpg"
        )

    dream(
        image_path=img_path,
        save_video=save_video,
        save_gif=save_gif,
        duration=duration,
        mirror_video=mirror_video,
    )


if __name__ == "__main__":
    main()



from dreamify.utils.common import (
    deprocess_image,
    get_image,
    preprocess_image,
    save_output,
    show,
)

__all__ = [show, preprocess_image, deprocess_image, get_image, save_output]



import tensorflow as tf


class DreamModel(tf.Module):
    def __init__(self, model):
        self.model = model

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
        )
    )
    def __call__(self, image, learning_rate, config):
        with tf.GradientTape() as tape:
            tape.watch(image)
            loss = DreamModel.compute_loss(image, config)
        grads = tape.gradient(loss, image)
        grads = tf.math.l2_normalize(grads)
        image += learning_rate * grads
        image = tf.clip_by_value(image, -1, 1)
        return loss, image

    def gradient_ascent_loop(
        self, image, iterations, learning_rate, max_loss=None, config=None
    ):
        for i in trange(
            iterations, desc="Gradient Ascent", unit="step", ncols=75, mininterval=0.1
        ):
            loss, image = gradient_ascent_step(image, learning_rate, config)

            if max_loss is not None and loss > max_loss:
                print(
                    f"\nTerminating early: Loss ({loss:.2f}) exceeded max_loss ({max_loss:.2f}).\n"
                )
                break

            framer = config.framer

            if config.enable_framing and framer.continue_framing():
                framer.add_to_frames(image)

        return image

    @staticmethod
    def compute_loss(input_image, feature_extractor):
        features = feature_extractor(input_image)
        loss = tf.zeros(shape=())
        for name in features.keys():
            coeff = feature_extractor.layer_settings[name]
            activation = features[name]
            loss += coeff * tf.reduce_mean(tf.square(activation[:, 2:-2, 2:-2, :]))
        return loss



from pathlib import Path

import pytest

from dreamify.dream import dream


@pytest.mark.filterwarnings(
    "ignore:The structure of `inputs` doesn't match the expected structure"
)
def test_mock_dream():
    img_path = "examples/example0.jpg"

    dream(
        img_path,
        output_path="mock_dream.png",
        octaves=1,
        iterations=5,
        learning_rate=15.0,
        save_video=True,
        save_gif=True,
        mirror_video=True,
    )
    Path("mock_dream.png").unlink(missing_ok=True)
    Path("mock_dream.mp4").unlink(missing_ok=True)
    Path("mock_dream.gif").unlink(missing_ok=True)


def test_dream_validator():
    img_path = "examples/example0.jpg"
    out_path = "examples/mock_dream.jpg"

    with pytest.raises(ValueError):
        dream(
            img_path,
            output_path=out_path,
            num_octave=-1,
            iterations=5,
            learning_rate=15.0,
        )
        print("Handled negative")
